@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2017-2019 Genome Research Ltd.
+ * Copyright (c) 2017-2020 Genome Research Ltd.
  * Author(s): James Bonfield
  *
  * Redistribution and use in source and binary forms, with or without
@@ -56,6 +56,8 @@
  * model.
  */
 
+#include "config.h"
+
 #include <stdint.h>
 #include <stdlib.h>
 #include <stdio.h>
@@ -74,6 +76,7 @@
 #include "varint.h"
 #include "pack.h"
 #include "rle.h"
+#include "utils.h"
 
 #define TF_SHIFT 12
 #define TOTFREQ (1<<TF_SHIFT)
@@ -97,76 +100,6 @@
  * are easier to understand, but can be up to 2x slower.
  */
 
-#define MAGIC 8
-
-#ifdef UNUSED
-static void hist1(unsigned char *in, unsigned int in_size, int F0[256]) {
-    int i;
-    for (i = 0; i < in_size; i++)
-	F0[in[i]]++;
-}
-
-static void hist4p(unsigned char *in, unsigned int in_size, int *F0) {
-    int F1[256+MAGIC] = {0}, F2[256+MAGIC] = {0}, F3[256+MAGIC] = {0};
-    int i;
-    unsigned char *in_end4 = in + (in_size & ~3);
-    unsigned char *in_end  = in + in_size;
-    while (in < in_end4) {
-	F0[*in++]++;
-	F1[*in++]++;
-	F2[*in++]++;
-	F3[*in++]++;
-    }
-    while (in < in_end)
-	F0[*in++]++;
-
-    for (i = 0; i < 256; i++)
-	F0[i] += F1[i] + F2[i] + F3[i];
-}
-#endif
-
-static void hist8(unsigned char *in, unsigned int in_size, int F0[256]) {
-    int F1[256+MAGIC] = {0}, F2[256+MAGIC] = {0}, F3[256+MAGIC] = {0};
-    int F4[256+MAGIC] = {0}, F5[256+MAGIC] = {0}, F6[256+MAGIC] = {0}, F7[256+MAGIC] = {0};
-    int i, i8 = in_size & ~7;
-    for (i = 0; i < i8; i+=8) {
-	F0[in[i+0]]++;
-	F1[in[i+1]]++;
-	F2[in[i+2]]++;
-	F3[in[i+3]]++;
-	F4[in[i+4]]++;
-	F5[in[i+5]]++;
-	F6[in[i+6]]++;
-	F7[in[i+7]]++;
-    }
-    while (i < in_size)
-	F0[in[i++]]++;
-
-    for (i = 0; i < 256; i++)
-	F0[i] += F1[i] + F2[i] + F3[i] + F4[i] + F5[i] + F6[i] + F7[i];
-}
-
-static void present8(unsigned char *in, unsigned int in_size, int F0[256]) {
-    int F1[256+MAGIC] = {0}, F2[256+MAGIC] = {0}, F3[256+MAGIC] = {0};
-    int F4[256+MAGIC] = {0}, F5[256+MAGIC] = {0}, F6[256+MAGIC] = {0}, F7[256+MAGIC] = {0};
-    int i, i8 = in_size & ~7;
-    for (i = 0; i < i8; i+=8) {
-	F0[in[i+0]]=1;
-	F1[in[i+1]]=1;
-	F2[in[i+2]]=1;
-	F3[in[i+3]]=1;
-	F4[in[i+4]]=1;
-	F5[in[i+5]]=1;
-	F6[in[i+6]]=1;
-	F7[in[i+7]]=1;
-    }
-    while (i < in_size)
-	F0[in[i++]]=1;
-
-    for (i = 0; i < 256; i++)
-	F0[i] += F1[i] + F2[i] + F3[i] + F4[i] + F5[i] + F6[i] + F7[i];
-}
-
 // Rounds to next power of 2.
 // credit to http://graphics.stanford.edu/~seander/bithacks.html
 static uint32_t round2(uint32_t v) {
@@ -180,7 +113,7 @@ static uint32_t round2(uint32_t v) {
     return v;
 }
 
-static int normalise_freq(int *F, int size, int tot) {
+static int normalise_freq(uint32_t *F, int size, uint32_t tot) {
     int m, M, j, loop = 0;
     uint64_t tr;
     if (!size)
@@ -232,7 +165,8 @@ static int normalise_freq(int *F, int size, int tot) {
 // A specialised version of normalise_freq_shift where the input size
 // is already normalised to a power of 2, meaning we can just perform
 // shifts instead of hard to define multiplications and adjustments.
-static void normalise_freq_shift(int *F, int size, int max_tot) {
+static void normalise_freq_shift(uint32_t *F, uint32_t size,
+				 uint32_t max_tot) {
     if (size == 0 || size == max_tot)
 	return;
 
@@ -245,7 +179,7 @@ static void normalise_freq_shift(int *F, int size, int max_tot) {
 }
 
 // symbols only
-static int encode_alphabet(uint8_t *cp, int *F) {
+static int encode_alphabet(uint8_t *cp, uint32_t *F) {
     uint8_t *op = cp;
     int rle, j;
 
@@ -271,7 +205,7 @@ static int encode_alphabet(uint8_t *cp, int *F) {
     return cp - op;
 }
 
-static int decode_alphabet(uint8_t *cp, uint8_t *cp_end, int *F) {
+static int decode_alphabet(uint8_t *cp, uint8_t *cp_end, uint32_t *F) {
     if (cp == cp_end)
 	return 0;
 
@@ -320,7 +254,7 @@ static int decode_alphabet(uint8_t *cp, uint8_t *cp_end, int *F) {
     return cp - op;
 }
 
-static int encode_freq(uint8_t *cp, int *F) {
+static int encode_freq(uint8_t *cp, uint32_t *F) {
     uint8_t *op = cp;
     int j;
 
@@ -334,7 +268,8 @@ static int encode_freq(uint8_t *cp, int *F) {
     return cp - op;
 }
 
-static int decode_freq(uint8_t *cp, uint8_t *cp_end, int *F, int *fsum) {
+static int decode_freq(uint8_t *cp, uint8_t *cp_end, uint32_t *F,
+		       uint32_t *fsum) {
     if (cp == cp_end)
 	return 0;
 
@@ -357,7 +292,7 @@ static int decode_freq(uint8_t *cp, uint8_t *cp_end, int *F, int *fsum) {
 // Use the order-0 freqs in F0 to encode the order-1 stats in F.
 // All symbols present in F are present in F0, but some in F0 will
 // be empty in F.  Thus we run-length encode the 0 frequencies.
-static int encode_freq_d(uint8_t *cp, int *F0, int *F) {
+static int encode_freq_d(uint8_t *cp, uint32_t *F0, uint32_t *F) {
     uint8_t *op = cp;
     int j, dz;
 
@@ -389,7 +324,8 @@ static int encode_freq_d(uint8_t *cp, int *F0, int *F) {
     return cp - op;
 }
 
-static int decode_freq_d(uint8_t *cp, uint8_t *cp_end, int *F0, int *F, int *total) {
+static int decode_freq_d(uint8_t *cp, uint8_t *cp_end, uint32_t *F0,
+			 uint32_t *F, uint32_t *total) {
     if (cp == cp_end)
 	return 0;
 
@@ -426,12 +362,13 @@ unsigned int rans_compress_bound_4x16(unsigned int size, int order) {
     if (!N) N=4;
 
     order &= 0xff;
-    return (order == 0
+    int sz = (order == 0
 	? 1.05*size + 257*3 + 4
 	: 1.05*size + 257*257*3 + 4 + 257*3+4) +
 	((order & X_PACK) ? 1 : 0) +
 	((order & X_RLE) ? 1 + 257*3+4: 0) + 20 +
 	((order & X_STRIPE) ? 1 + 5*N: 0);
+    return sz + (sz&1) + 2; // make this even so buffers are word aligned
 }
 
 // Compresses in_size bytes from 'in' to *out_size bytes in 'out'.
@@ -448,7 +385,8 @@ unsigned char *rans_compress_O0_4x16(unsigned char *in, unsigned int in_size,
     RansState rans1;
     RansState rans3;
     uint8_t* ptr;
-    int F[256+MAGIC] = {0}, i, j, tab_size = 0, rle, x;
+    uint32_t F[256+MAGIC] = {0};
+    int i, j, tab_size = 0, rle, x;
     int bound = rans_compress_bound_4x16(in_size,0)-20; // -20 for order/size/meta
 
     if (!out) {
@@ -458,6 +396,10 @@ unsigned char *rans_compress_O0_4x16(unsigned char *in, unsigned int in_size,
     if (!out || bound > *out_size)
 	return NULL;
 
+    // If "out" isn't word aligned, tweak out_end/ptr to ensure it is.
+    // We already added more round in bound to allow for this.
+    if (((size_t)out)&1)
+	bound--;
     ptr = out_end = out + bound;
 
     if (in_size == 0)
@@ -467,8 +409,8 @@ unsigned char *rans_compress_O0_4x16(unsigned char *in, unsigned int in_size,
     hist8(in, in_size, F);
 
     // Normalise so frequences sum to power of 2
-    int fsum = in_size;
-    int max_val = round2(fsum);
+    uint32_t fsum = in_size;
+    uint32_t max_val = round2(fsum);
     if (max_val > TOTFREQ)
 	max_val = TOTFREQ;
 
@@ -584,7 +526,7 @@ unsigned char *rans_uncompress_O0_4x16(unsigned char *in, unsigned int in_size,
 	return NULL;
 
     // Precompute reverse lookup of frequency.
-    int F[256] = {0}, fsum;
+    uint32_t F[256] = {0}, fsum;
     int fsz = decode_freq(cp, cp_end, F, &fsum);
     if (!fsz)
 	goto err;
@@ -595,7 +537,7 @@ unsigned char *rans_uncompress_O0_4x16(unsigned char *in, unsigned int in_size,
     // Build symbols; fixme, do as part of decode, see the _d variant
     for (j = x = 0; j < 256; j++) {
 	if (F[j]) {
-	    if (x + F[j] > TOTFREQ || F[j] < 0)
+	    if (F[j] > TOTFREQ - x)
 		goto err;
 	    for (y = 0; y < F[j]; y++) {
 		ssym [y + x] = j;
@@ -673,66 +615,6 @@ unsigned char *rans_uncompress_O0_4x16(unsigned char *in, unsigned int in_size,
     return NULL;
 }
 
-#ifdef UNUSED
-static void hist1_1(unsigned char *in, unsigned int in_size,
-		    int F0[256][256], int T0[256]) {
-    unsigned int last_i, i;
-    unsigned char c;
-
-    for (last_i=i=0; i<in_size; i++) {
-	F0[last_i][c = in[i]]++;
-	T0[last_i]++;
-	last_i = c;
-    }
-}
-#endif
-
-static void hist1_4(unsigned char *in, unsigned int in_size,
-		    int F0[256][256], int *T0) {
-    int T1[256+MAGIC] = {0}, T2[256+MAGIC] = {0}, T3[256+MAGIC] = {0};
-    unsigned int idiv4 = in_size/4;
-    int i;
-    unsigned char c0=0, c1=0, c2=0, c3=0;
-
-    unsigned char *in0 = in + 0;
-    unsigned char *in1 = in + idiv4;
-    unsigned char *in2 = in + idiv4*2;
-    unsigned char *in3 = in + idiv4*3;
-
-    unsigned char last_0 = 0, last_1 = in1[-1], last_2 = in2[-1], last_3 = in3[-1];
-    //unsigned char last_0 = 0, last_1 = 0, last_2 = 0, last_3 = 0;
-
-    unsigned char *in0_end = in1;
-
-    while (in0 < in0_end) {
-	F0[last_0][c0 = *in0++]++;
-	T0[last_0]++;
-	last_0 = c0;
-
-	F0[last_1][c1 = *in1++]++;
-	T1[last_1]++;
-	last_1 = c1;
-
-	F0[last_2][c2 = *in2++]++;
-	T2[last_2]++;
-	last_2 = c2;
-
-	F0[last_3][c3 = *in3++]++;
-	T3[last_3]++;
-	last_3 = c3;
-    }
-
-    while (in3 < in + in_size) {
-	F0[last_3][c3 = *in3++]++;
-	T3[last_3]++;
-	last_3 = c3;
-    }
-
-    for (i = 0; i < 256; i++) {
-	T0[i]+=T1[i]+T2[i]+T3[i];
-    }
-}
-
 //-----------------------------------------------------------------------------
 
 double fast_log(double a) {
@@ -744,7 +626,8 @@ double fast_log(double a) {
 // 10 bit means smaller memory footprint when decoding and
 // more speed due to cache hits, but it *may* be a poor
 // compression fit.
-static int compute_shift(int *F0, int (*F)[256], int *T, int *S) {
+static int compute_shift(uint32_t *F0, uint32_t (*F)[256], uint32_t *T,
+			 int *S) {
     int i, j;
 
     double e10 = 0, e12 = 0;
@@ -823,9 +706,12 @@ unsigned char *rans_compress_O1_4x16(unsigned char *in, unsigned int in_size,
     if (!out || bound > *out_size)
 	return NULL;
 
+    if (((size_t)out)&1)
+	bound--;
     out_end = out + bound;
 
-    int F[256][256] = {{0}}, T[256+MAGIC] = {0}, i, j;
+    uint32_t F[256][256] = {{0}}, T[256+MAGIC] = {0};
+    int i, j;
 
     //memset(F, 0, 256*256*sizeof(int));
     //memset(T, 0, 256*sizeof(int));
@@ -840,7 +726,7 @@ unsigned char *rans_compress_O1_4x16(unsigned char *in, unsigned int in_size,
     *cp++ = 0; // uncompressed header marker
 
     // Encode the order-0 symbols for use in the order-1 frequency tables
-    int F0[256+MAGIC] = {0};
+    uint32_t F0[256+MAGIC] = {0};
     present8(in, in_size, F0);
     F0[0]=1;
     cp += encode_alphabet(cp, F0);
@@ -869,7 +755,7 @@ unsigned char *rans_compress_O1_4x16(unsigned char *in, unsigned int in_size,
 
 	normalise_freq_shift(F[i], T[i], 1<<shift); T[i]=1<<shift;
 
-	int *F_i_ = F[i];
+	uint32_t *F_i_ = F[i];
 	for (x = j = 0; j < 256; j++) {
 	    RansEncSymbolInit(&syms[i][j], x, F_i_[j], shift);
 	    x += F_i_[j];
@@ -1069,7 +955,7 @@ unsigned char *rans_uncompress_O1_4x16(unsigned char *in, unsigned int in_size,
     }
 
     // Decode order-0 symbol list; avoids needing in order-1 tables
-    int F0[256] = {0};
+    uint32_t F0[256] = {0};
     int fsz = decode_alphabet(cp, c_freq_end, F0);
     if (!fsz)
 	goto err;
@@ -1082,7 +968,7 @@ unsigned char *rans_uncompress_O1_4x16(unsigned char *in, unsigned int in_size,
 	if (F0[i] == 0)
 	    continue;
 
-	int F[256] = {0}, T = 0;
+	uint32_t F[256] = {0}, T = 0;
 	fsz = decode_freq_d(cp, c_freq_end, F0, F, &T);
 	if (!fsz)
 	    goto err;
@@ -1098,7 +984,7 @@ unsigned char *rans_uncompress_O1_4x16(unsigned char *in, unsigned int in_size,
 	// Build symbols; fixme, do as part of decode, see the _d variant
 	for (j = x = 0; j < 256; j++) {
 	    if (F[j]) {
-		if (x + F[j] > (1<<shift) || F[j] < 0)
+		if (F[j] > (1<<shift) - x)
 		    goto err;
 
 		memset(&sfb[i][x], j, F[j]);
@@ -1463,52 +1349,6 @@ unsigned char *rans_compress_4x16(unsigned char *in, unsigned int in_size,
     return rans_compress_to_4x16(in, in_size, NULL, out_size, order);
 }
 
-/*
- * Data transpose by N.
- * Tuned for specific common cases of N.
- */
-static void unstripe(unsigned char *out, unsigned char *outN,
-		     unsigned int ulen, unsigned int N,
-		     unsigned int idxN[256]) {
-    int j = 0, k;
-
-    switch (N) {
-    case 4:
-	while (j < ulen-4) {
-	    for (k = 0; k < 4; k++)
-		out[j++] = outN[idxN[k]++];
-	}
-	while (j < ulen) {
-	    for (k = 0; j < ulen; k++)
-		out[j++] = outN[idxN[k]++];
-	}
-	break;
-
-    case 2:
-	while (j < ulen-4) {
-	    for (k = 0; k < 2; k++)
-		out[j++] = outN[idxN[k]++];
-	}
-	while (j < ulen) {
-	    for (k = 0; j < ulen; k++)
-		out[j++] = outN[idxN[k]++];
-	}
-	break;
-
-    default:
-	// General case, around 25% slower overall decode
-	while (j < ulen-N) {
-	    for (k = 0; k < N; k++)
-		out[j++] = outN[idxN[k]++];
-	}
-	while (j < ulen) {
-	    for (k = 0; j < ulen; k++)
-		out[j++] = outN[idxN[k]++];
-	}
-	break;
-    }
-}
-
 unsigned char *rans_uncompress_to_4x16(unsigned char *in,  unsigned int in_size,
 				       unsigned char *out, unsigned int *out_size) {
     unsigned char *in_end = in + in_size;
@@ -1524,6 +1364,8 @@ unsigned char *rans_uncompress_to_4x16(unsigned char *in,  unsigned int in_size,
 
 	// Decode lengths
 	c_meta_len += var_get_u32(in+c_meta_len, in_end, &ulen);
+	if (c_meta_len >= in_size)
+	    return NULL;
 	unsigned int N = in[c_meta_len++];
 	unsigned int clenN[256], ulenN[256], idxN[256];
 	if (!out) {
@@ -1755,6 +1597,8 @@ unsigned char *rans_uncompress_to_4x16(unsigned char *in,  unsigned int in_size,
 
     if (do_rle) {
 	// Unpack RLE.  tmp1 -> tmp2.
+	if (u_meta_size == 0)
+	    goto err;
 	uint64_t unrle_size = *out_size;
 	int rle_nsyms = *meta ? *meta : 256;
 	if (u_meta_size < 1+rle_nsyms)
