@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2016-2019 Genome Research Ltd.
+ * Copyright (c) 2016-2020 Genome Research Ltd.
  * Author(s): James Bonfield
  * 
  * Redistribution and use in source and binary forms, with or without 
@@ -83,7 +83,7 @@
 //   @VP2-06:112:H7LNDMCVY:1:1105:27100:1172 1:N:0:ATTCAGAA+AGGCGAAG
 //   @VP2-06:112:H7LNDMCVY:1:1105:27172:1172 1:N:0:ATTCAGAA+AGGCTAAG
 
-
+#include "config.h"
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -194,6 +194,14 @@ static name_context *create_context(int max_names) {
 	return NULL;
 #endif
 
+    // An arbitrary limit to prevent malformed data from consuming excessive
+    // amounts of memory.  Consider upping this if we have genuine use cases
+    // for larger blocks.
+    if (max_names > 1e7) {
+	fprintf(stderr, "Name codec currently has a max of 10 million rec.\n");
+	return NULL;
+    }
+
 #ifndef NO_THREADS
     pthread_once(&tok_once, tok_tls_init);
 
@@ -203,7 +211,7 @@ static name_context *create_context(int max_names) {
 	if (!ctx) return NULL;
 	ctx->max_names = max_names;
 	pthread_setspecific(tok_key, ctx);
-    } else if (ctx->max_names < max_names) {
+    } else if (ctx->max_names < max_names+1) {
 	ctx = realloc(ctx, sizeof(*ctx) + ++max_names*sizeof(*ctx->lc));
 	if (!ctx) return NULL;
 	ctx->max_names = max_names;
@@ -226,6 +234,8 @@ static name_context *create_context(int max_names) {
      memset(&ctx->token_icount[0], 0, sizeof(int));
      ctx->max_tok = 1;
 
+     ctx->lc[0].last_ntok = 0;
+
     return ctx;
 }
 
@@ -362,8 +372,11 @@ static int encode_token_int(name_context *ctx, int ntok,
     if (encode_token_type(ctx, ntok, type) < 0) return -1;
     if (descriptor_grow(&ctx->desc[id], 4) < 0)	return -1;
 
-    // Assumes little endian and unalign access OK.
-    *(uint32_t *)(ctx->desc[id].buf + ctx->desc[id].buf_l) = val;
+    uint8_t *cp = &ctx->desc[id].buf[ctx->desc[id].buf_l];
+    cp[0] = (val >>  0) & 0xff;
+    cp[1] = (val >>  8) & 0xff;
+    cp[2] = (val >> 16) & 0xff;
+    cp[3] = (val >> 24) & 0xff;
     ctx->desc[id].buf_l += 4;
 
     return 0;
@@ -377,8 +390,8 @@ static int decode_token_int(name_context *ctx, int ntok,
     if (ctx->desc[id].buf_l + 4 > ctx->desc[id].buf_a)
 	return -1;
 
-    // Assumes little endian and unalign access OK.
-    *val = *(uint32_t *)(ctx->desc[id].buf + ctx->desc[id].buf_l);
+    uint8_t *cp = ctx->desc[id].buf + ctx->desc[id].buf_l;
+    *val = (cp[0]) + (cp[1]<<8) + (cp[2]<<16) + ((uint32_t)cp[3]<<24);
     ctx->desc[id].buf_l += 4;
 
     return 0;
@@ -1045,6 +1058,8 @@ static int decode_name(name_context *ctx, char *name, int name_len) {
 	tok = decode_token_type(ctx, ntok);
 	//fprintf(stderr, "Tok %d = %d\n", ntok, tok);
 
+	ctx->lc[cnum].last_ntok = 0;
+
 	switch (tok) {
 	case N_CHAR:
 	    if (len+1 >= name_len) return -1;
@@ -1076,6 +1091,7 @@ static int decode_name(name_context *ctx, char *name, int name_len) {
 	    break;
 
 	case N_DDELTA0:
+	    if (ntok >= ctx->lc[pnum].last_ntok) return -1;
 	    if (decode_token_int1(ctx, ntok, N_DDELTA0, &v) < 0) return -1;
 	    v += ctx->lc[pnum].last_token_int[ntok];
 	    if (len+ctx->lc[pnum].last_token_str[ntok]+1 >= name_len) return -1;
@@ -1096,6 +1112,7 @@ static int decode_name(name_context *ctx, char *name, int name_len) {
 	    break;
 
 	case N_DDELTA:
+	    if (ntok >= ctx->lc[pnum].last_ntok) return -1;
 	    if (decode_token_int1(ctx, ntok, N_DDELTA, &v) < 0) return -1;
 	    v += ctx->lc[pnum].last_token_int[ntok];
 	    if (len+20 >= name_len) return -1;
@@ -1110,6 +1127,7 @@ static int decode_name(name_context *ctx, char *name, int name_len) {
 	    break;
 
 	case N_MATCH:
+	    if (ntok >= ctx->lc[pnum].last_ntok) return -1;
 	    switch (ctx->lc[pnum].last_token_type[ntok]) {
 	    case N_CHAR:
 		if (len+1 >= name_len) return -1;
@@ -1478,12 +1496,18 @@ uint8_t *encode_names(char *blk, int len, int level, int use_arith,
     }
 
     uint8_t *cp = out;
-    //*out_len = tot_size+4;
-    //*(uint32_t *)cp = tot_size;   cp += 4;
 
     *out_len = tot_size;
-    *(uint32_t *)cp = last_start; cp += 4;
-    *(uint32_t *)cp = nreads;     cp += 4;
+//    *(uint32_t *)cp = last_start; cp += 4;
+//    *(uint32_t *)cp = nreads;     cp += 4;
+    *cp++ = (last_start >>  0) & 0xff;
+    *cp++ = (last_start >>  8) & 0xff;
+    *cp++ = (last_start >> 16) & 0xff;
+    *cp++ = (last_start >> 24) & 0xff;
+    *cp++ = (nreads     >>  0) & 0xff;
+    *cp++ = (nreads     >>  8) & 0xff;
+    *cp++ = (nreads     >> 16) & 0xff;
+    *cp++ = (nreads     >> 24) & 0xff;
     *cp++ = use_arith;
     //write(1, &nreads, 4);
     int last_tnum = -1;
@@ -1524,12 +1548,15 @@ uint8_t *decode_names(uint8_t *in, uint32_t sz, uint32_t *out_len) {
 	return NULL;
 
     int i, o = 9;
-    int ulen   = *(uint32_t *)in;
+    //int ulen   = *(uint32_t *)in;
+    int ulen   = (in[0]<<0) | (in[1]<<8) | (in[2]<<16) |
+	(((uint32_t)in[3])<<24);
 
     if (ulen < 0 || ulen >= INT_MAX-1024)
 	return NULL;
 
-    int nreads = *(uint32_t *)(in+4);
+    //int nreads = *(uint32_t *)(in+4);
+    int nreads = (in[4]<<0) | (in[5]<<8) | (in[6]<<16) | (((uint32_t)in[7])<<24);
     int use_arith = in[8];
     name_context *ctx = create_context(nreads);
     if (!ctx)
@@ -1552,6 +1579,7 @@ uint8_t *decode_names(uint8_t *in, uint32_t sz, uint32_t *out_len) {
 	    }
 
 	    if ((ttype & 15) != 0 && (ttype & 128)) {
+		if (tnum < 0) goto err;
 		ctx->desc[tnum<<4].buf = malloc(nreads);
 		if (!ctx->desc[tnum<<4].buf)
 		    goto err;
@@ -1562,6 +1590,7 @@ uint8_t *decode_names(uint8_t *in, uint32_t sz, uint32_t *out_len) {
 		memset(&ctx->desc[tnum<<4].buf[1], N_MATCH, nreads-1);
 	    }
 
+	    if (tnum < 0) goto err;
 	    i = (tnum<<4) | (ttype&15);
 	    if (j >= i)
 		goto err;
@@ -1588,6 +1617,7 @@ uint8_t *decode_names(uint8_t *in, uint32_t sz, uint32_t *out_len) {
 	}
 
 	if ((ttype & 15) != 0 && (ttype & 128)) {
+	    if (tnum < 0) goto err;
 	    if (ctx->desc[tnum<<4].buf) free(ctx->desc[tnum<<4].buf);
 	    ctx->desc[tnum<<4].buf = malloc(nreads);
 	    if (!ctx->desc[tnum<<4].buf)
@@ -1604,6 +1634,7 @@ uint8_t *decode_names(uint8_t *in, uint32_t sz, uint32_t *out_len) {
 	int64_t clen, ulen = uncompressed_size(&in[o], sz-o);
 	if (ulen < 0 || ulen >= INT_MAX)
 	    goto err;
+	if (tnum < 0) goto err;
 	i = (tnum<<4) | (ttype&15);
 
 	if (i >= MAX_TBLOCKS || i < 0)
